/* -*- Mode: C; c-file-style: "stroustrup" -*- */

/* NATools - Package with parallel corpora tools
 * Copyright (C) 1998-2001  Djoerd Hiemstra
 * Copyright (C) 2002-2006  Alberto Simes
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <signal.h>
/* #include <bits/signum.h> */
#include <time.h>
#include <stdlib.h>



#include <glib.h>

#include "parseini.h"
#include "corpusinfo.h"

#define DONE(fd)	write(fd, "** DONE **\n", 11)
#define ERROR(fd)       write(fd, "** SYNTAX ERROR **\n", 19)

#define DEBUG 0

int sockfd, fd;

GHashTable *CORPORA;
gint32 LAST_CORPORA = 0;

void LOG(char* log, ...) {
    char stime[80];
    time_t timep;

    va_list args;
    va_start(args, log);

    time(&timep);
    strftime(stime, 80, "%F %T", localtime(&timep));

    fprintf(stderr, "[%s] ", stime);
    vfprintf(stderr, log, args);
    fprintf(stderr,"\n");
    va_end(args);
}

#if DEBUG
static guint32 length(const guint32 *needle) {
    guint32 ans = 0;
    while(needle[ans]) { ans++; }
    return ans;
}
#endif


void parse(int fd, char* buf, char* dir);

static int dump_corpora_list(int fd, gint last, GHashTable *corpora) {
    gboolean toexit = FALSE;
    gchar sbuf[1024];
    gint i;

    sprintf(sbuf, "%d\n", last);
    if (write(fd, sbuf, strlen(sbuf)) < 0) 
	toexit = TRUE;

    for (i = 1; i <= last && !toexit; i++) {
	sprintf(sbuf, "[%d] %s\n", 
		i,
		(char*)g_hash_table_lookup(((CorpusInfo*)g_hash_table_lookup(corpora, &i))->config,"name"));
	if (write(fd, sbuf, strlen(sbuf)) < 0) 
	    toexit = TRUE;
    }
    return toexit;
}

static int dump_conf(CorpusInfo *corpus, int fd, char *key) {
    gchar sbuf[1024];
    gboolean toexit = FALSE;

    sprintf(sbuf, "%s\n", (char*)g_hash_table_lookup(corpus->config, key));
    if (write(fd, sbuf, strlen(sbuf)) < 0) 
	toexit = TRUE;

    return toexit;
}

static CorpusInfo* get_corpus(gint id) {
    if (id > LAST_CORPORA)
	return NULL;
    return (CorpusInfo*)g_hash_table_lookup(CORPORA, &id);
}

double* rank_load(CorpusInfo *corpus, const char* file, const guint32 size) {
    FILE *rank;

    if (corpus->rank_cache1 && strcmp(corpus->rank_cache_filename1, file) == 0) {
	corpus->has_rank = 1;
	return corpus->rank_cache1;
    } else if (corpus->rank_cache2 && strcmp(corpus->rank_cache_filename2, file) == 0) {
	corpus->has_rank = 1;
	return corpus->rank_cache2;
    } else {
	rank = fopen(file, "rb");
	if (rank) {
	    double *ranks;
	    corpus->has_rank = 1;

	    ranks = g_new(double, size);
	    fread(ranks, sizeof(double), size, rank);
	    fclose(rank);

	    if (corpus->last_rank_cache == 1) {
		corpus->last_rank_cache = 2;
		if (corpus->rank_cache1) {
		    g_free(corpus->rank_cache1);
		    g_free(corpus->rank_cache_filename1);
		}
		corpus->rank_cache1 = ranks;
		corpus->rank_cache_filename1 = strdup(file);
		return corpus->rank_cache1;
	    } else {
		corpus->last_rank_cache = 1;
		if (corpus->rank_cache2) {
		    g_free(corpus->rank_cache2);
		    g_free(corpus->rank_cache_filename2);
		}
		corpus->rank_cache2 = ranks;
		corpus->rank_cache_filename2 = strdup(file);
		return corpus->rank_cache2;
	    }
	} else {
	    corpus->has_rank = 0;
	    return NULL;
	}
    }
}

CorpusCell *corpus_retrieve_sentence(CorpusInfo* corpus,
				     gboolean source,
				     const guchar chunk,
				     guint32 sentence,
				     double *kwalitee) 
{
    guint32 begin, end, delta;
    gchar *rank_filename;
    guint32 *offsets;
    double *ranks = NULL;
    guint32 size;
    FILE *fh;
    CorpusCell *buf;

    char *basedir = g_hash_table_lookup(corpus->config, "homedir");

    size = corpus->chunks[chunk-1].size;
    if (source) {
	offsets = corpus->chunks[chunk-1].source_offset;
        fh = corpus->chunks[chunk-1].source_crp;
    } else {
	offsets = corpus->chunks[chunk-1].target_offset;
        fh = corpus->chunks[chunk-1].target_crp;
    }

    rank_filename = g_strdup_printf("%s/rank.%03hhu.rnk", basedir, chunk);
    ranks = rank_load(corpus, rank_filename, size);

    begin = offsets[sentence];
    end = offsets[sentence+1];
    delta = end - begin;

    buf = g_new0(CorpusCell, delta);
    fseek(fh, begin*sizeof(CorpusCell) + CORPUS_HEADER_SIZE, 0);
    fread( buf, sizeof(CorpusCell),delta,fh);

    if (corpus->has_rank) {
	*kwalitee = ranks[sentence];
    }

    return buf;
}


static int send_kwalitee(CorpusInfo *corpus, int sd, double q)
{
    char sbuf[1024];
    gboolean toexit = FALSE;

    if (!corpus->has_rank) return toexit;

    sprintf(sbuf, "%% %f\n", q);
    if (write(sd, sbuf, strlen(sbuf)) < 0)
	toexit = 1;

    return toexit;
}

static int corpus_send_sentence(int sd,
				WordLstNode **W,
				CorpusCell *buf)
{
    char sbuf[1024];
    gboolean toexit = FALSE;
    CorpusCell *ptr;
    
    ptr = buf;
    while(!toexit && ptr->word) {
	char *word = word_list_get_by_id(W, ptr->word);
	if (ptr->flags & 0x1) {
	    word = uppercase_dup(word);
	} else if (ptr->flags & 0x2) {
	    word = capital_dup(word);
	} else {
	    word = g_strdup(word);
	}

	sprintf(sbuf, "%s ", word);
	if (write(sd, sbuf, strlen(sbuf)) < 0)
	    toexit = 1;
	g_free(word);

	ptr++;
    }
    write(sd, "\n", 1);
    return toexit;
}

void dump_dict(int fd, char* word, gint32 dir, CorpusInfo* crp) {
    guint32 j;
    char sbuf[300];
    guint32 wid, twid;
    float prob;

    WordList *S;
    WordLstNode **T;
    Dictionary *D;

    if (strcmp(word, "")==0) { DONE(fd); return; }

    if (dir > 0) {
        S = crp->SourceLex;
        T = crp->TargetLexIds;
        D = crp->SourceTarget;
    } else {
        S = crp->TargetLex;
        T = crp->SourceLexIds;
        D = crp->TargetSource;
    }
    
    wid = word_list_get_id(S, word);
    if (!wid) { ERROR(fd); return; }

    sprintf(sbuf, "%u\n", dictionary_get_occ(D, wid));
    write(fd, sbuf, strlen(sbuf));

    for (j = 0; j < MAXENTRY; j++) {
	twid = 0;
	prob = 0.0;
	twid = dictionary_get_id(D, wid, j);
	if (twid) {
	    prob = dictionary_get_val(D, wid, j);
	    sprintf(sbuf, "%.6f %s\n", prob, word_list_get_by_id(T, twid));
	    write(fd, sbuf, strlen(sbuf));
	}
    }
    
    DONE(fd);

    return;
}


void handle_sigpipe(int x) {
    LOG("Connection abruptely closed");
}

void handle_sigalrm(int x) {
    LOG("Timeout, dude!");
    close(fd);
}

void handle_sigint(int x) {
    LOG("Quiting...");
    close(fd);
    close(sockfd);
    exit(0);
}

int* int_ptr(gint x) {
    gint *y;
    y = g_new(gint, 1);
    *y = x;
    return y;
}

int main(int argc, char *argv[]) {
    struct sockaddr_in serv;
    struct sockaddr_in client;
    //    struct in_addr *aux;

    short port = 4000;

    //    char *host = "193.136.19.131";

    char buf[1024];
    size_t n;
    unsigned int size = sizeof(client);
    int zbr = 1;
    FILE *f = NULL;

    if (argc != 2) {
	printf("Usage: nat-server <config-file>\n");
	return 0;
    }

    CORPORA = g_hash_table_new(g_int_hash, g_int_equal);

    /* CONFIGURATION FILE */
    LOG("Loading configuration file");
    f = fopen(argv[1], "rb");
    if (f) {
	CorpusInfo *tmp_corpus;
	while(!feof(f)) {
	    fgets(buf, 1024, f);
	    if (!feof(f)) {
		if (buf[0] == '\n' || buf[0] == '#' || buf[0] == ' ') continue;
		chomp(buf);
		LOG("Loading corpus from %s [%d]", buf, ++LAST_CORPORA);
		tmp_corpus = corpus_info_new(buf);
		g_hash_table_insert(CORPORA, int_ptr(LAST_CORPORA), tmp_corpus);
	    }
	}
	fclose(f);
    } else {
	report_error("Can't find '%s'", argv[1]);
    }
    
    /* SERVER CODE */
    /*-------------*/

    signal(SIGALRM, handle_sigalrm);
    signal(SIGINT, handle_sigint);
    signal(SIGPIPE, handle_sigpipe);

    memset(&serv, 0, sizeof(serv));

    serv.sin_family = AF_INET;
    serv.sin_port = htons(port);
    serv.sin_addr.s_addr = htonl(INADDR_ANY);
 
    if (( sockfd = socket(AF_INET, SOCK_STREAM, 0) ) < 0 ) {
	LOG("couldn't create socket!");
	return -1;
    } else {
	LOG("socket created!");
    }

    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &zbr, sizeof zbr);

    if ((bind(sockfd, (struct sockaddr *) &serv, sizeof(serv))) <0 ) {
	LOG("couldn't bind!");
	return -1;
    } else {
	LOG("bind!");
    }

  
    if (listen(sockfd,5) < 0 ) {
	LOG("couldn't listen");
	return -1;
    } else{
	LOG("listen done");
    }

    for(;;) {
	if ((fd = accept(sockfd, (struct sockaddr *) &client, &size)) < 0 ) {
	    LOG("couln't accept connection!");
	    return(-1);
	} else {
#if DEBUG	    
	    LOG("accept connection (from: %s:%d) (fd: %d)",
		inet_ntoa((struct in_addr)client.sin_addr),
		ntohs(client.sin_port),
		fd); 
#endif
	    alarm(50);
	    n = read(fd, buf, 1024);
	    if ( n > 0 )
		parse(fd, buf, argv[1]);
	    alarm(0);
	    close(fd);  
	}
    }
}

void play(int fd) {
    char buff[1000];
    strcpy(buff, "<html>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "<body>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "<table width='100%' height='100%'>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "<tr><td align='center' valign='middle'>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "<span style='color: #ff0000; padding: 20px; font-size: xx-large; border: solid 1px #ff0000'>I am logging you!</span>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "</td></tr>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "</table>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "</body>\n");
    write(fd, buff, strlen(buff));
    strcpy(buff, "</html>\n");
    write(fd, buff, strlen(buff));


}

void parse(int fd, char* buf, char* dir)
{
    /* word word */

    CorpusInfo *corpus = NULL;

    int direction = 0;
    gboolean exact_match = FALSE;
    gboolean need_free = FALSE;
    gboolean both = FALSE;
    char words[50][150];
    guint32 wids[50];
    int i = 0, j = 2;
    char *token = NULL;
    guint32 wid;
    guint32 *occs = NULL, *occs1 = NULL, *occs2 = NULL;
    guint32 counter, total = 20;
    guint32 *otherwids = NULL;
		
    chomp(buf);

    for (i=0;i<50;i++) strcpy(words[i],"");
    i = 0;

    if (strcmp(buf, "") == 0) return;

#if DEBUG
    LOG("Request was [%s]", buf);
#endif

    token = strtok(buf, " ");
    while(token) {
	strcpy(words[i], token);
	i++;
	token = strtok(NULL, " ");
    }

    if (strncmp(words[0], "LIST", 4) == 0) {
	dump_corpora_list(fd, LAST_CORPORA, CORPORA);
	return;
    } else if (strncmp(words[0], "?", 1) == 0) {
	dump_conf(get_corpus(atoi(words[1])), fd, words[2]);
	return;
    } else if (strncmp(words[0], "~>", 2) == 0) {
	corpus = get_corpus(atoi(words[1]));
	dump_dict(fd, words[2], 1, corpus);
	return;
    } else if (strncmp(words[0], "<~", 2) == 0) {
	corpus = get_corpus(atoi(words[1]));
	dump_dict(fd, words[2], -1, corpus);
	return;
    } else if (strncmp(words[0], "<->", 3) == 0) {
	direction = 1;
	both = TRUE;
    } else if (strncmp(words[0], "<=>", 3) == 0) {
	direction = 1;
	both = TRUE;
	exact_match = TRUE;
    } else if (strncmp(words[0], "<-", 2) == 0) {
	direction = -1;
    } else if (strncmp(words[0], "->", 2) == 0) {
	direction = 1;
    } else if (strncmp(words[0], "<=", 2) == 0) {
	direction = -1;
	exact_match = TRUE;
    } else if (strncmp(words[0], "=>", 2) == 0) {
	direction = 1;
	exact_match = TRUE;
    } else if (strncmp(words[0], "GET", 3) == 0) {
	LOG("Playing http server");
	play(fd);
	return;
    }

    corpus = get_corpus(atoi(words[1]));
    
    if (!corpus) {
	ERROR(fd);
    } else if (!direction) {
	ERROR(fd);
    } else {

	if (strcmp(words[j], "<=>") == 0 || strcmp(words[j], "<->") == 0) {
	    ERROR(fd);
	    return;
	}

	if (strcmp(words[j], "*") == 0) {
	    wids[j-2] = 1;
	} else {
	    wid = word_list_get_id( (direction>0)?corpus->SourceLex:corpus->TargetLex, words[j]);

	    if (!wid) {
		ERROR(fd);
		return;
	    }

	    wids[j-2] = wid;
	
	    occs = inv_index_compact_get_occurrences( (direction > 0)?corpus->SourceIdx:corpus->TargetIdx, wid);
	}
	wids[j] = 0;

	
	if (i > 3) {
	    for (j = 3; j < i; j++) {

		if (words[j][0] == '#' && isdigit(words[j][1])) {
		    sscanf(words[j],"#%d", &total);
		    total = total<0?20:(total>100?100:total);
		    wids[j-2] = 0;
		    break;
		} else if (strcmp(words[j], "<=>") == 0 || strcmp(words[j], "<->") == 0) {
		    if (direction < 0) {
			ERROR(fd);
			return;
		    }
		    direction = -direction;
		    wids[j-2] = 0;
		    otherwids = &(wids[j]);
		} else {
		    if (strcmp(words[j], "*") == 0) {
			wids[j-2] = 1;
		    } else {
			wid = word_list_get_id( (direction > 0)?corpus->SourceLex:corpus->TargetLex, words[j]);
			
			if (!wid) {
			    if (need_free) g_free(occs);
			    DONE(fd);
			    return;
			}

			wids[j-2] = wid;
		    
			occs1 = inv_index_compact_get_occurrences(
                            (direction > 0)?corpus->SourceIdx:corpus->TargetIdx,
                            wid);
			occs2 = intersect(occs, occs1);
			if (need_free) g_free(occs);
			occs = occs2;
			need_free = TRUE;
		    }
		}
	    }
	    wids[j-2] = 0;
	}
	
#if DEBUG      
	LOG("Retrieving and processing %d unities", length(occs));
#endif
	
	counter = 0;
	
	occs1 = occs;
	
	while(occs1 && *occs && counter < total) {
	    guchar  chunk;
	    double q1 = -1, q2 = -1;
	    gboolean has_kwalitee = FALSE;
	    guint32 sentence = unpack(*occs, &chunk);
	    CorpusCell *src, *trg;

	    src = corpus_retrieve_sentence(corpus,  TRUE, chunk, sentence-1, &q1);
	    trg = corpus_retrieve_sentence(corpus, FALSE, chunk, sentence-1, &q2);

	    if (q1 >= 0 && q2 >= 0) has_kwalitee = TRUE;
		
	    if (exact_match && both) {
		if (corpus_strstr(src, wids) &&
		    corpus_strstr(trg, otherwids)) {
		    if (send_kwalitee(corpus, fd, q1)) break;
		    if (corpus_send_sentence(fd, corpus->SourceLexIds, src)) break;
		    if (corpus_send_sentence(fd, corpus->TargetLexIds, trg)) break;
		    counter ++;
		}
	    } else if (both) {
		    if (send_kwalitee(corpus, fd, q1)) break;
		if (corpus_send_sentence(fd, corpus->SourceLexIds, src)) break;
		if (corpus_send_sentence(fd, corpus->TargetLexIds, trg)) break;
		counter++;
	    } else if (direction < 0) {
		if (!exact_match || (exact_match && corpus_strstr(trg, wids))) {
		    if (send_kwalitee(corpus, fd, q1)) break;
		    if (corpus_send_sentence(fd, corpus->SourceLexIds, src)) break;
		    if (corpus_send_sentence(fd, corpus->TargetLexIds, trg)) break;
		    counter++;
		}
	    } else {
		if (!exact_match || (exact_match && corpus_strstr(src, wids))) {
		    if (send_kwalitee(corpus, fd, q1)) break;
		    if (corpus_send_sentence(fd, corpus->SourceLexIds, src)) break;
		    if (corpus_send_sentence(fd, corpus->TargetLexIds, trg)) break;
		    counter++;
		}
	    }

	    g_free(src);
	    g_free(trg);
	    
	    occs++;
	}

#if DEBUG
	LOG("Sent %d units", counter); 
#endif
	DONE(fd);
	
	if (need_free) g_free(occs1);
    } 
}

